//Fuzz Click (Special FX) beeper engine port from ZX Spectrum to Arduino
//the original Z80 code is by Jonathan Smith
//port by Shiru (shiru@mail.ru) 26.02.17


//include music data generated by the bin2.py script

#include "music_data.h"

//include sample data

#include "drum_sample_data.h"



//general settings

#define COMPILE_ADR           40000             //address for compiled Beepola module, default is 40000

#define SPEAKER_PORT          PORTD             //speaker is on the port D
#define SPEAKER_DDR           DDRD
#define SPEAKER_BIT           (1<<7)            //PD7 (Uno pin 7)
#define SPEAKER_BITN           (1<<2)            //PD2 (Uno pin 2)

//there is no sample rate, width of the output slots in AVR clocks used instead
//calculated for 22988 Hz sample rate, which is closest to the original Tritone 22875 Hz

#define LOOP_HALF_TIME        384     //AVR cycles that matches half of the loop time of the original engine
#define PULSE_UNIT_TIME       72      //how many AVR cycles per one pulse unit

#define SAMPLE_RATE           (16000000/(LOOP_HALF_TIME*2))     //just for reference; drums played at doubled sample rate, 41666 Hz

#define FRAME_TIME            64      //in sample rate units



//array of counters and reload values, for each channel

unsigned char ch_enabled[2];
unsigned char ch_count  [2];
unsigned char ch_load   [2];
unsigned char ch_pulse_a[2];
unsigned char ch_pulse_b[2];
unsigned char ch_decay_cnt[2];
unsigned char ch_decay_set[2];

unsigned char timer_value_a;
unsigned char timer_value_b;

//drum sample player variables

unsigned char drum_request;
unsigned char drum_output;
unsigned char drum_pulse_length;

const unsigned char *drum_data;


//sync counter to syncronize the music parser with the synth code
//it is volatile because it the main thread reads it back

volatile unsigned int parser_sync;


//song parser variables

unsigned char *order_list_tone;
unsigned char *order_list_perc;

unsigned char *pattern_ptr_tone_1;
unsigned char *pattern_ptr_tone_2;
unsigned char *pattern_ptr_perc;

unsigned char ch_delay[3];



//pointer to the actual interrupt handler that gets changed all the time

void (*interrupt_handler)(void);



//frequency dividers for the original engine timings

const unsigned char freq_table[] PROGMEM ={
	0xFD,
	0xEE,0xE1,0xD4,0xC8,0xBD,0xB2,0xA8,0x9F,0x96,0x8E,0x86,0x7E,
	0x77,0x70,0x6A,0x64,0x5E,0x59,0x54,0x4F,0x4B,0x47,0x43,0x3F,
	0x3B,0x38,0x35,0x32,0x2F,0x2C,0x2A,0x27,0x25,0x23,0x21,0x1F,
	0x1D,0x1C,0x1B,0x19,0x17,0x16,0x15,0x13,0x12,0x11,0x10,0x0F,
	0x0E,0x0D,0x0C,0x01,0x00
};



//initialize all

void setup()
{
	unsigned char chn;
	Serial.begin(9600);
	cli();

pinMode(4,OUTPUT);
digitalWrite(4,HIGH);
	//initialize channel variables

	for(chn=0;chn<2;++chn)
	{
		ch_enabled[chn]=0;
		ch_count  [chn]=0;
		ch_load   [chn]=0;
		ch_pulse_a[chn]=1;
		ch_pulse_b[chn]=0;
		ch_decay_cnt[chn]=0;
		ch_decay_set[chn]=0;
	}

	drum_request=0;

	//initialize song parser variables

	parser_sync=0;

	order_list_perc=music_data+((pgm_read_byte_near(music_data)+(pgm_read_byte_near(music_data+1)<<8))-COMPILE_ADR);
	order_list_tone=music_data+2;

	song_new_pattern();

	//set a port pin as the output

	SPEAKER_DDR|=SPEAKER_BIT; 

	//set timer2 to generate interrupts at the sample rate

	TCCR2A=0;
	TCCR2B=0;
	TCNT2 =0;

	TCCR2A|=(1<<WGM21);
	TCCR2B|=(0<<CS02)|(1<<CS01)|(1<<CS00);  //prescaler=32, not enough max value for the pulse length otherwise
	TIMSK2|=(1<<OCIE2A);

	//force the timer setting and next interrupt handler to start the handler sequence

	interrupt_handler_main_loop_part_1();

	sei();
}




void interrupt_handler_main_loop_part_1(void)
{
	OCR2A=LOOP_HALF_TIME/32-1;

	if(drum_request)  //only change the handler into drum mode inside an interrupt, as it is not an atom operation (two bytes)
	{
		drum_request=0;
		interrupt_handler=interrupt_handler_drum;
		return;
	}

	--ch_count[0];

	if(!ch_count[0])
	{
		ch_count[0]=ch_load[0];

		timer_value_a=(1+ch_pulse_a[0])*(PULSE_UNIT_TIME/32)-1;
		timer_value_b=(1+ch_pulse_b[0])*(PULSE_UNIT_TIME/32)-1;
		
		interrupt_handler=interrupt_handler_pulse_1_in;
	}
	else
	{
		interrupt_handler=interrupt_handler_main_loop_part_2;
	}
}



void interrupt_handler_main_loop_part_2(void)
{
	OCR2A=LOOP_HALF_TIME/32-1;

	if(drum_request)
	{
		drum_request=0;
		interrupt_handler=interrupt_handler_drum;
		return;
	}

	--ch_count[1];

	if(parser_sync) --parser_sync;

	if(!ch_count[1])
	{
		ch_count[1]=ch_load[1];

		timer_value_a=(1+ch_pulse_a[1])*(PULSE_UNIT_TIME/32)-1;
		timer_value_b=(1+ch_pulse_b[1])*(PULSE_UNIT_TIME/32)-1;
		
		interrupt_handler=interrupt_handler_pulse_2_in;
	}
	else
	{
		interrupt_handler=interrupt_handler_main_loop_part_1;
	}
}



void interrupt_handler_pulse_1_in(void)
{
	OCR2A=timer_value_a;

	if(ch_enabled[0]) SPEAKER_PORT=SPEAKER_BIT;

	interrupt_handler=interrupt_handler_pulse_1_out;
}



void interrupt_handler_pulse_1_out(void)
{
	OCR2A=timer_value_b;

	SPEAKER_PORT=0;

	interrupt_handler=interrupt_handler_main_loop_part_2;
}



void interrupt_handler_pulse_2_in(void)
{
	OCR2A=timer_value_a;

	if(ch_enabled[1]) SPEAKER_PORT=SPEAKER_BIT;

	interrupt_handler=interrupt_handler_pulse_2_out;
}



void interrupt_handler_pulse_2_out(void)
{
	OCR2A=timer_value_b;

	SPEAKER_PORT=0;

	interrupt_handler=interrupt_handler_main_loop_part_1;
}



//drum sample player interrupt

void interrupt_handler_drum(void)
{
	OCR2A=LOOP_HALF_TIME/32-1;

	SPEAKER_PORT=drum_output;

	if(drum_pulse_length)
	{
		--drum_pulse_length;
	}
	else
	{
		drum_output^=SPEAKER_BIT;

		drum_pulse_length=pgm_read_byte_near(drum_data);

		++drum_data;
		
		if(!drum_pulse_length) interrupt_handler=interrupt_handler_main_loop_part_1;
	}

	if(parser_sync) --parser_sync; //drum sample rate is twice higher than tone sample rate, keep sync
}



ISR(TIMER2_COMPA_vect)
{
	interrupt_handler();    //call currently assigned interrupt handler using function pointer
}



void song_new_pattern(void)
{
	unsigned int off;

	while(1)
	{
		off=pgm_read_byte_near(order_list_tone)+(pgm_read_byte_near(order_list_tone+1)<<8);

		order_list_tone+=2;
		
		if(off) break;

		off=pgm_read_byte_near(order_list_tone)+(pgm_read_byte_near(order_list_tone+1)<<8);

		order_list_tone=music_data+(off-COMPILE_ADR);
	}

	pattern_ptr_tone_1=music_data+(off-COMPILE_ADR);
	pattern_ptr_tone_2=music_data+(pgm_read_byte_near(pattern_ptr_tone_1)+(pgm_read_byte_near(pattern_ptr_tone_1+1)<<8)-COMPILE_ADR);

	pattern_ptr_tone_1+=2;

	while(1)
	{
		off=pgm_read_byte_near(order_list_perc)+(pgm_read_byte_near(order_list_perc+1)<<8);

		order_list_perc+=2;
		
		if(off) break;

		off=pgm_read_byte_near(order_list_perc)+(pgm_read_byte_near(order_list_perc+1)<<8);

		order_list_perc=music_data+(off-COMPILE_ADR);
	}

	pattern_ptr_perc=music_data+(off-COMPILE_ADR);

	ch_delay[0]=0;
	ch_delay[1]=0;
	ch_delay[2]=0;
}



//main loop

void loop()
{
	unsigned char chn,tag,wait;

	while(1)
	{
		//update row

		if(ch_delay[0]) --ch_delay[0];
		
		if(!ch_delay[0])
		{
			while(1)
			{
				tag=pgm_read_byte_near(pattern_ptr_tone_1);

				if(tag==0x80)  //end of pattern
				{
					song_new_pattern();
					
					continue;
				}

				if(tag==0x81)  //sustain value
				{
					ch_decay_set[0]=pgm_read_byte_near(pattern_ptr_tone_1+1);

					pattern_ptr_tone_1+=2;

					continue;
				}

				break;
			}

			ch_delay[0]=pgm_read_byte_near(pattern_ptr_tone_1+1);
			
			pattern_ptr_tone_1+=2;

			if(tag<0x80)
			{
				ch_load   [0]=pgm_read_byte_near(freq_table+tag);
				ch_enabled[0]=1;
			}
			else
			{
				ch_load   [0]=0x40;
				ch_enabled[0]=0;
			}

			ch_pulse_a  [0]=ch_load[0]>>3;
			ch_pulse_b  [0]=1;
			ch_decay_cnt[0]=ch_decay_set[0];
		}

		if(ch_delay[1]) --ch_delay[1];

		if(!ch_delay[1])
		{
			while(1)
			{
				tag=pgm_read_byte_near(pattern_ptr_tone_2);

				if(tag==0x80)  //end of pattern
				{
					//do nothing, patterns always has the same duration
				}

				if(tag==0x81)  //sustain value
				{
					ch_decay_set[1]=pgm_read_byte_near(pattern_ptr_tone_2+1);

					pattern_ptr_tone_2+=2;

					continue;
				}

				break;
			}
			
			ch_delay[1]=pgm_read_byte_near(pattern_ptr_tone_2+1);
			
			pattern_ptr_tone_2+=2;

			if(tag<0x80)
			{
				ch_load   [1]=pgm_read_byte_near(freq_table+tag);
				ch_enabled[1]=1;
			}
			else
			{
				ch_load   [1]=0x40;
				ch_enabled[1]=0;
			}

			ch_pulse_a  [1]=ch_load[1]>>2;
			ch_pulse_b  [1]=1;
			ch_decay_cnt[1]=ch_decay_set[1];
		}

		if(ch_delay[2]) --ch_delay[2];

		if(!ch_delay[2])
		{
			while(1)
			{
				tag=pgm_read_byte_near(pattern_ptr_perc);

				if(tag==0x80)  //end of pattern
				{
					//do nothing, patterns always has the same duration
				}

				break;
			}
			
			ch_delay[2]=pgm_read_byte_near(pattern_ptr_perc+1);
			
			pattern_ptr_perc+=2;

			if(tag>=0x81&&tag<0x81+4)
			{
				drum_data=(const unsigned char*)pgm_read_word(&(drum_sample_list[tag-0x81]));
				
				drum_output=0;
				drum_pulse_length=0;
				drum_request=1;       //interrupt handler will get changed in the next tone interrupt
			}
		}
		
		// run decay envelopes

		for(chn=0;chn<2;++chn)
		{
			--ch_decay_cnt[chn];

			if(!ch_decay_cnt[chn])
			{
				ch_decay_cnt[chn]=ch_decay_set[chn];

				if(ch_pulse_a[chn]>1)
				{
					--ch_pulse_a[chn];
					++ch_pulse_b[chn];
				}
			}
		}
		
		parser_sync=SAMPLE_RATE/50;

		//wait for the next row
		//delay 1 is important, by some reason song tempo starts to jump a lot without it

		while(parser_sync>0) delay(1);
	}

}
